# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import argparse
import numpy as np
import os
import random

import horovod.torch as hvd
import torch

from ofa.imagenet_classification.elastic_nn.modules.dynamic_op import DynamicSeparableConv2d
from ofa.imagenet_classification.elastic_nn.networks import OFAMobileNetV3
from ofa.imagenet_classification.run_manager import DistributedImageNetRunConfig
from ofa.imagenet_classification.networks import MobileNetV3Large
from ofa.imagenet_classification.run_manager.distributed_run_manager import DistributedRunManager
from ofa.utils import download_url, MyRandomResizedCrop
from ofa.imagenet_classification.elastic_nn.training.progressive_shrinking import load_models
import wandb

import warnings
warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
#parser.add_argument('--task', type=str, default='depth', choices=[
#    'large', 'kernel', 'depth', 'expand',
#])
parser.add_argument('--name', type=str, help='job name', default='image_classification')
parser.add_argument('--bs', type=int, help='batch size per worker (GPU)', default=16)
parser.add_argument('--lr', type=float, help='initial learning rate', default=0.03)

args = parser.parse_args()

def set_task_params(args):
    if args.task == 'large':
        args.path = 'exp/ic/%s/s1' % args.name
        args.dynamic_batch_size = 1
        args.n_epochs = 1 #10
        args.base_lr = args.lr
        args.warmup_epochs = 0 #5
        args.warmup_lr = -1
        args.ks_list = '7'
        args.expand_list = '6'
        args.depth_list = '4'
    elif args.task == 'kernel':
        args.path = 'exp/ic/%s/s2' % args.name
        args.dynamic_batch_size = 1
        args.n_epochs = 1 #5
        args.base_lr = args.lr / 2.0
        args.warmup_epochs = 0 #1
        args.warmup_lr = -1
        args.ks_list = '3,5,7'
        args.expand_list = '6'
        args.depth_list = '4'
    elif args.task == 'depth':
        args.path = 'exp/ic/%s/s3' % args.name
        args.dynamic_batch_size = 2
        args.n_epochs = 1 #5
        args.base_lr = args.lr / 4.0
        args.warmup_epochs = 0 #1
        args.warmup_lr = -1
        args.ks_list = '3,5,7'
        args.expand_list = '6'
        args.depth_list = '2,3,4'
    elif args.task == 'expand':
        args.path = 'exp/ic/%s/s4' % args.name
        args.dynamic_batch_size = 4
        args.n_epochs = 1 #5
        args.base_lr = args.lr / 12.0
        args.warmup_epochs = 0 #1
        args.warmup_lr = -1
        args.ks_list = '3,5,7'
        args.expand_list = '3,4,6'
        args.depth_list = '2,3,4'
    else:
        raise NotImplementedError

args.manual_seed = 0

args.lr_schedule_type = 'cosine'

args.base_batch_size = args.bs
args.valid_size = 10000

args.opt_type = 'sgd'
args.momentum = 0.9
args.no_nesterov = False
args.weight_decay = 3e-5
args.label_smoothing = 0.1
args.no_decay_keys = 'bn#bias'
args.fp16_allreduce = False

args.model_init = 'he_fout'
args.validation_frequency = 1
args.print_frequency = 10

args.n_worker = 8
args.resize_scale = 0.08
args.distort_color = 'tf'
args.image_size = '128,160,192,224'
args.continuous_size = True
args.not_sync_distributed_image_size = False

args.bn_momentum = 0.1
args.bn_eps = 1e-5
args.dropout = 0.1
args.base_stage_width = 'proxyless'

args.width_mult_list = '1.0'
args.dy_conv_scaling_mode = 1
args.independent_distributed_sampling = False

args.kd_ratio = 0.0
args.kd_type = 'ce'


if __name__ == '__main__':

    # Initialize Horovod
    hvd.init()
    print(hvd.local_rank())
    # Pin GPU to be used to process local rank (one GPU per process)
    torch.cuda.set_device(hvd.local_rank())

    num_gpus = hvd.size()

    # init wandb
    if hvd.rank() == 0:
        wandb.init(project='dist-automl', entity='automl2022', name="logs/wandb_%s_class.log"%args.name, config=args)

    torch.manual_seed(args.manual_seed)
    torch.cuda.manual_seed_all(args.manual_seed)
    np.random.seed(args.manual_seed)
    random.seed(args.manual_seed)

    # image size
    args.image_size = [int(img_size) for img_size in args.image_size.split(',')]
    if len(args.image_size) == 1:
        args.image_size = args.image_size[0]
    MyRandomResizedCrop.CONTINUOUS = args.continuous_size
    MyRandomResizedCrop.SYNC_DISTRIBUTED = not args.not_sync_distributed_image_size

    # build run config from args
    args.lr_schedule_param = None
    args.opt_param = {
        'momentum': args.momentum,
        'nesterov': not args.no_nesterov,
    }
    args.width_mult_list = [float(width_mult) for width_mult in args.width_mult_list.split(',')]
    args.width_mult_list = args.width_mult_list[0] if len(args.width_mult_list) == 1 else args.width_mult_list

    # training
    from ofa.imagenet_classification.elastic_nn.training.progressive_shrinking import validate, train

    tasks = ['large', 'kernel', 'depth', 'expand']
    for task in tasks:
        args.task = task
        set_task_params(args)
        os.makedirs(args.path, exist_ok=True)

        args.init_lr = args.base_lr * num_gpus * (args.base_batch_size / 64.0) # linearly rescale the learning rate
        if args.warmup_lr < 0:
            args.warmup_lr = args.base_lr
        args.train_batch_size = args.base_batch_size
        args.test_batch_size = args.base_batch_size * 2
        run_config = DistributedImageNetRunConfig(**args.__dict__, num_replicas=num_gpus, rank=hvd.rank())

        # print run config information
        if hvd.rank() == 0:
            print('Run config:')
            for k, v in run_config.config.items():
                print('\t%s: %s' % (k, v))

        if args.dy_conv_scaling_mode == -1:
            args.dy_conv_scaling_mode = None
        DynamicSeparableConv2d.KERNEL_TRANSFORM_MODE = args.dy_conv_scaling_mode

        # build net from args
        args.ks_list = [int(ks) for ks in args.ks_list.split(',')]
        args.expand_list = [int(e) for e in args.expand_list.split(',')]
        args.depth_list = [int(d) for d in args.depth_list.split(',')]

        net = OFAMobileNetV3(
            n_classes=run_config.data_provider.n_classes, bn_param=(args.bn_momentum, args.bn_eps),
            dropout_rate=args.dropout, base_stage_width=args.base_stage_width, width_mult=args.width_mult_list,
            ks_list=args.ks_list, expand_ratio_list=args.expand_list, depth_list=args.depth_list
        )

        """ Distributed RunManager """
        # Horovod: (optional) compression algorithm.
        compression = hvd.Compression.fp16 if args.fp16_allreduce else hvd.Compression.none
        distributed_run_manager = DistributedRunManager(
            args.path, net, run_config, compression, args.name, backward_steps=args.dynamic_batch_size, is_root=(hvd.rank() == 0)
        )
        distributed_run_manager.save_config()
        # hvd broadcast
        distributed_run_manager.broadcast()

        validate_func_dict = {'image_size_list': {224} if isinstance(args.image_size, int) else sorted({160, 224}),
                              'ks_list': sorted({min(args.ks_list), max(args.ks_list)}),
                              'expand_ratio_list': sorted({min(args.expand_list), max(args.expand_list)}),
                              'depth_list': sorted({min(net.depth_list), max(net.depth_list)})}


        if args.task == 'large' or args.task == 'kernel':
            validate_func_dict['ks_list'] = sorted(args.ks_list)
            if args.task == 'kernel':
                args.ofa_checkpoint_path = 'exp/ic/%s/s1/checkpoint/model_best.pth.tar' % args.name
                load_models(distributed_run_manager, distributed_run_manager.net, args.ofa_checkpoint_path)
                distributed_run_manager.write_log(
                    '%.3f\t%.3f\t%.3f\t%s' % validate(distributed_run_manager, is_test=True, **validate_func_dict), 'valid')
            train(distributed_run_manager, args,
                  lambda _run_manager, epoch, is_test: validate(_run_manager, epoch, is_test, **validate_func_dict))
        elif args.task == 'depth':
            from ofa.imagenet_classification.elastic_nn.training.progressive_shrinking import train_elastic_depth
            args.ofa_checkpoint_path = 'exp/ic/%s/s2/checkpoint/model_best.pth.tar' % args.name
            train_elastic_depth(train, distributed_run_manager, args, validate_func_dict)
        elif args.task == 'expand':
            from ofa.imagenet_classification.elastic_nn.training.progressive_shrinking import train_elastic_expand
            args.ofa_checkpoint_path = 'exp/ic/%s/s3/checkpoint/model_best.pth.tar' % args.name
            train_elastic_expand(train, distributed_run_manager, args, validate_func_dict)
        else:
            raise NotImplementedError
